@@ -49,11 +49,14 @@
 
 #include "mersenne.h"
 #include "half.h"
+#include "bfloat16.h"
 
 #include "cub/util_debug.cuh"
 #include "cub/util_device.cuh"
 #include "cub/util_type.cuh"
 #include "cub/util_macro.cuh"
+#include "cub/util_math.cuh"
+#include "cub/util_ptx.cuh"
 #include "cub/iterator/discard_output_iterator.cuh"
 
 /******************************************************************************
@@ -80,8 +83,23 @@ T SafeBitCast(const U& in)
 /**
  * Assert equals
  */
-#define AssertEquals(a, b) if ((a) != (b)) { std::cerr << "\n(" << __FILE__ << ": " << __LINE__ << ")\n"; exit(1);}
+#define AssertEquals(a, b)                                                     \
+  if ((a) != (b))                                                              \
+  {                                                                            \
+    std::cerr << "\n"                                                          \
+              << __FILE__ << ": " << __LINE__                                  \
+              << ": AssertEquals(" #a ", " #b ") failed.\n";                   \
+    exit(1);                                                                   \
+  }
 
+#define AssertTrue(a)                                                          \
+  if (!(a))                                                                    \
+  {                                                                            \
+    std::cerr << "\n"                                                          \
+              << __FILE__ << ": " << __LINE__                                  \
+              << ": AssertTrue(" #a ") failed.\n";                             \
+    exit(1);                                                                   \
+  }
 
 /******************************************************************************
  * Command-line parsing functionality
@@ -286,7 +304,7 @@ struct CommandLineArgs
             CubDebugExit(cudaMemGetInfo(&device_free_physmem, &device_total_physmem));
 
             int ptx_version = 0;
-            error = CubDebug(cub::PtxVersion(ptx_version));
+            error = CubDebug(CUB_NS_QUALIFIER::PtxVersion(ptx_version));
             if (error) break;
 
             error = CubDebug(cudaGetDeviceProperties(&deviceProp, dev));
@@ -405,7 +423,15 @@ __noinline__ bool IsNaN<half_t>(half_t val)
         ((bits >= 0xFC01) /*&& (bits <= 0xFFFFFFFF)*/));
 }
 
+template<>
+__noinline__ bool IsNaN<bfloat16_t>(bfloat16_t val)
+{
+    const auto bits = SafeBitCast<unsigned short>(val);
 
+    // commented bit is always true, leaving for documentation:
+    return (((bits >= 0x7F81) && (bits <= 0x7FFF)) ||
+        ((bits >= 0xFF81) /*&& (bits <= 0xFFFFFFFF)*/));
+}
 
 /**
  * Generates random keys.
@@ -521,37 +547,91 @@ int CoutCast(signed char val) { return val; }
  */
 enum GenMode
 {
-    UNIFORM,            // Assign to '2', regardless of integer seed
-    INTEGER_SEED,       // Assign to integer seed
-    RANDOM,             // Assign to random, regardless of integer seed
-    RANDOM_BIT,         // Assign to randomly chosen 0 or 1, regardless of integer seed
+    UNIFORM,                 // Assign to '2', regardless of integer seed
+    INTEGER_SEED,            // Assign to integer seed
+    RANDOM,                  // Assign to random, regardless of integer seed
+    RANDOM_BIT,              // Assign to randomly chosen 0 or 1, regardless of integer seed
+    RANDOM_MINUS_PLUS_ZERO,  // Assign to random, with some values being -0.0 or +0.0 patterns
 };
 
 /**
  * Initialize value
  */
+#pragma nv_exec_check_disable
 template <typename T>
 __host__ __device__ __forceinline__ void InitValue(GenMode gen_mode, T &value, int index = 0)
 {
-    switch (gen_mode)
+    // RandomBits is host-only.
+    if (CUB_IS_DEVICE_CODE)
     {
-#if (CUB_PTX_ARCH == 0)
-    case RANDOM:
-        RandomBits(value);
-        break;
-    case RANDOM_BIT:
-        char c;
-        RandomBits(c, 0, 0, 1);
-        value = (c > 0) ? (T) 1 : (T) -1;
-        break;
-#endif
-     case UNIFORM:
-        value = 2;
-        break;
-    case INTEGER_SEED:
-    default:
-         value = (T) index;
-        break;
+        #if CUB_INCLUDE_DEVICE_CODE
+            switch (gen_mode)
+            {
+                case RANDOM:
+                case RANDOM_BIT:
+                case RANDOM_MINUS_PLUS_ZERO:
+                    _CubLog("%s\n",
+                            "cub::InitValue cannot generate random numbers on device.");
+                    CUB_NS_QUALIFIER::ThreadTrap();
+                    break;
+                case UNIFORM:
+                    value = 2;
+                    break;
+                case INTEGER_SEED:
+                default:
+                    value = (T) index;
+                    break;
+            }
+        #endif // CUB_INCLUDE_DEVICE_CODE
+    }
+    else
+    {
+        #if CUB_INCLUDE_HOST_CODE
+            switch (gen_mode)
+            {
+                case RANDOM:
+                    RandomBits(value);
+                    break;
+                case RANDOM_BIT:
+                {
+                    char c;
+                    RandomBits(c, 0, 0, 1);
+                    value = (c > 0) ? (T) 1 : (T) -1;
+                    break;
+                }
+                case RANDOM_MINUS_PLUS_ZERO:
+                {
+                    // Replace roughly 1/128 of values with -0.0 or +0.0, and generate the rest randomly
+                    typedef typename CUB_NS_QUALIFIER::Traits<T>::UnsignedBits UnsignedBits;
+                    char c;
+                    RandomBits(c);
+                    if (c == 0)
+                    {
+                        // Replace 1/256 of values with +0.0 bit pattern
+                        value = SafeBitCast<T>(UnsignedBits(0));
+                    }
+                    else if (c == 1)
+                    {
+                        // Replace 1/256 of values with -0.0 bit pattern
+                        value = SafeBitCast<T>(UnsignedBits(UnsignedBits(1) <<
+                                                            (sizeof(UnsignedBits) * 8) - 1));
+                    }
+                    else
+                    {
+                        // 127/128 of values are random
+                        RandomBits(value);
+                    }
+                    break;
+                }
+                case UNIFORM:
+                    value = 2;
+                    break;
+                case INTEGER_SEED:
+                default:
+                    value = (T) index;
+                    break;
+            }
+        #endif // CUB_INCLUDE_HOST_CODE
     }
 }
 
@@ -559,25 +639,53 @@ __host__ __device__ __forceinline__ void InitValue(GenMode gen_mode, T &value, i
 /**
  * Initialize value (bool)
  */
+#pragma nv_exec_check_disable
 __host__ __device__ __forceinline__ void InitValue(GenMode gen_mode, bool &value, int index = 0)
 {
-    switch (gen_mode)
+    // RandomBits is host-only.
+    if (CUB_IS_DEVICE_CODE)
     {
-#if (CUB_PTX_ARCH == 0)
-    case RANDOM:
-    case RANDOM_BIT:
-        char c;
-        RandomBits(c, 0, 0, 1);
-        value = (c > 0);
-        break;
-#endif
-     case UNIFORM:
-        value = true;
-        break;
-    case INTEGER_SEED:
-    default:
-        value = (index > 0);
-        break;
+        #if CUB_INCLUDE_DEVICE_CODE
+            switch (gen_mode)
+            {
+                case RANDOM:
+                case RANDOM_BIT:
+                case RANDOM_MINUS_PLUS_ZERO:
+                    _CubLog("%s\n",
+                            "cub::InitValue cannot generate random numbers on device.");
+                    CUB_NS_QUALIFIER::ThreadTrap();
+                    break;
+                case UNIFORM:
+                    value = true;
+                    break;
+                case INTEGER_SEED:
+                default:
+                    value = (index > 0);
+                    break;
+            }
+        #endif // CUB_INCLUDE_DEVICE_CODE
+    }
+    else
+    {
+        #if CUB_INCLUDE_HOST_CODE
+            switch (gen_mode)
+            {
+                case RANDOM:
+                case RANDOM_BIT:
+                case RANDOM_MINUS_PLUS_ZERO:
+                    char c;
+                    RandomBits(c, 0, 0, 1);
+                    value = (c > 0);
+                    break;
+                case UNIFORM:
+                    value = true;
+                    break;
+                case INTEGER_SEED:
+                default:
+                    value = (index > 0);
+                    break;
+            }
+        #endif // CUB_INCLUDE_HOST_CODE
     }
 }
 
@@ -586,7 +694,7 @@ __host__ __device__ __forceinline__ void InitValue(GenMode gen_mode, bool &value
  * cub::NullType test initialization
  */
 __host__ __device__ __forceinline__ void InitValue(GenMode /* gen_mode */,
-						   cub::NullType &/* value */,
+						   CUB_NS_QUALIFIER::NullType &/* value */,
 						   int /* index */ = 0)
 {}
 
@@ -594,17 +702,34 @@ __host__ __device__ __forceinline__ void InitValue(GenMode /* gen_mode */,
 /**
  * cub::KeyValuePair<OffsetT, ValueT>test initialization
  */
+#pragma nv_exec_check_disable
 template <typename KeyT, typename ValueT>
 __host__ __device__ __forceinline__ void InitValue(
     GenMode                             gen_mode,
-    cub::KeyValuePair<KeyT, ValueT>&    value,
+    CUB_NS_QUALIFIER::KeyValuePair<KeyT, ValueT>&    value,
     int                                 index = 0)
 {
     InitValue(gen_mode, value.value, index);
 
-    // Assign corresponding flag with a likelihood of the last bit being set with entropy-reduction level 3
-    RandomBits(value.key, 3);
-    value.key = (value.key & 0x1);
+    // This specialization only appears to be used by test_warp_scan.
+    // It initializes with uniform values and random keys, so we need to
+    // protect the call to the host-only RandomBits.
+    if (CUB_IS_DEVICE_CODE)
+    {
+        #if CUB_INCLUDE_DEVICE_CODE
+            _CubLog("%s\n",
+                    "cub::InitValue cannot generate random numbers on device.");
+            CUB_NS_QUALIFIER::ThreadTrap();
+        #endif // CUB_INCLUDE_DEVICE_CODE
+    }
+    else
+    {
+        #if CUB_INCLUDE_HOST_CODE
+            // Assign corresponding flag with a likelihood of the last bit being set with entropy-reduction level 3
+            RandomBits(value.key, 3);
+            value.key = (value.key & 0x1);
+        #endif // CUB_INCLUDE_HOST_CODE
+    }
 }
 
 
@@ -617,7 +742,7 @@ __host__ __device__ __forceinline__ void InitValue(
  * KeyValuePair ostream operator
  */
 template <typename Key, typename Value>
-std::ostream& operator<<(std::ostream& os, const cub::KeyValuePair<Key, Value> &val)
+std::ostream& operator<<(std::ostream& os, const CUB_NS_QUALIFIER::KeyValuePair<Key, Value> &val)
 {
     os << '(' << CoutCast(val.key) << ',' << CoutCast(val.value) << ')';
     return os;
@@ -681,7 +806,7 @@ std::ostream& operator<<(std::ostream& os, const cub::KeyValuePair<Key, Value> &
         T retval = make_##T(a.x + b.x);                     \
         return retval;                                      \
     }                                                       \
-    namespace cub {                                         \
+    CUB_NAMESPACE_BEGIN                                     \
     template<>                                              \
     struct NumericTraits<T>                                 \
     {                                                       \
@@ -703,7 +828,7 @@ std::ostream& operator<<(std::ostream& os, const cub::KeyValuePair<Key, Value> &
             return retval;                                  \
         }                                                   \
     };                                                      \
-    } /* namespace std */
+    CUB_NAMESPACE_END
 
 
 
@@ -769,7 +894,7 @@ std::ostream& operator<<(std::ostream& os, const cub::KeyValuePair<Key, Value> &
             a.y + b.y);                                     \
         return retval;                                      \
     }                                                       \
-    namespace cub {                                         \
+    CUB_NAMESPACE_BEGIN                                         \
     template<>                                              \
     struct NumericTraits<T>                                 \
     {                                                       \
@@ -793,7 +918,7 @@ std::ostream& operator<<(std::ostream& os, const cub::KeyValuePair<Key, Value> &
             return retval;                                  \
         }                                                   \
     };                                                      \
-    } /* namespace cub */
+    CUB_NAMESPACE_END
 
 
 
@@ -866,7 +991,7 @@ std::ostream& operator<<(std::ostream& os, const cub::KeyValuePair<Key, Value> &
             a.z + b.z);                                     \
         return retval;                                      \
     }                                                       \
-    namespace cub {                                         \
+    CUB_NAMESPACE_BEGIN                                     \
     template<>                                              \
     struct NumericTraits<T>                                 \
     {                                                       \
@@ -892,7 +1017,7 @@ std::ostream& operator<<(std::ostream& os, const cub::KeyValuePair<Key, Value> &
             return retval;                                  \
         }                                                   \
     };                                                      \
-    } /* namespace cub */
+    CUB_NAMESPACE_END
 
 
 /**
@@ -971,7 +1096,7 @@ std::ostream& operator<<(std::ostream& os, const cub::KeyValuePair<Key, Value> &
             a.w + b.w);                                     \
         return retval;                                      \
     }                                                       \
-    namespace cub {                                         \
+    CUB_NAMESPACE_BEGIN                                     \
     template<>                                              \
     struct NumericTraits<T>                                 \
     {                                                       \
@@ -999,7 +1124,7 @@ std::ostream& operator<<(std::ostream& os, const cub::KeyValuePair<Key, Value> &
             return retval;                                  \
         }                                                   \
     };                                                      \
-    } /* namespace cub */
+    CUB_NAMESPACE_END
 
 /**
  * All vector overloads
@@ -1036,10 +1161,15 @@ CUB_VEC_OVERLOAD(double, double)
  */
 struct TestFoo
 {
-    long long   x;
-    int         y;
-    short       z;
-    char        w;
+    using x_t = long long;
+    using y_t = int;
+    using z_t = short;
+    using w_t = char;
+
+    x_t x;
+    y_t y;
+    z_t z;
+    w_t w;
 
     // Factory
     static __host__ __device__ __forceinline__ TestFoo MakeTestFoo(long long x, int y, short z, char w)
@@ -1051,10 +1181,10 @@ struct TestFoo
     // Assignment from int operator
     __host__ __device__ __forceinline__ TestFoo& operator =(int b)
     {
-        x = b;
-        y = b;
-        z = b;
-        w = b;
+        x = static_cast<x_t>(b);
+        y = static_cast<y_t>(b);
+        z = static_cast<z_t>(b);
+        w = static_cast<w_t>(b);
         return *this;
     }
 
@@ -1118,7 +1248,7 @@ __host__ __device__ __forceinline__ void InitValue(GenMode gen_mode, TestFoo &va
 
 
 /// numeric_limits<TestFoo> specialization
-namespace cub {
+CUB_NAMESPACE_BEGIN
 template<>
 struct NumericTraits<TestFoo>
 {
@@ -1145,7 +1275,7 @@ struct NumericTraits<TestFoo>
             NumericTraits<char>::Lowest());
     }
 };
-} // namespace cub
+CUB_NAMESPACE_END
 
 
 //---------------------------------------------------------------------
@@ -1234,7 +1364,7 @@ __host__ __device__ __forceinline__ void InitValue(GenMode gen_mode, TestBar &va
 }
 
 /// numeric_limits<TestBar> specialization
-namespace cub {
+CUB_NAMESPACE_BEGIN
 template<>
 struct NumericTraits<TestBar>
 {
@@ -1257,7 +1387,7 @@ struct NumericTraits<TestBar>
             NumericTraits<int>::Lowest());
     }
 };
-} // namespace cub
+CUB_NAMESPACE_END
 
 
 /******************************************************************************
@@ -1315,7 +1445,7 @@ int CompareResults(float* computed, float* reference, OffsetT len, bool verbose
  * Compares the equivalence of two arrays
  */
 template <typename OffsetT>
-int CompareResults(cub::NullType* computed, cub::NullType* reference, OffsetT len, bool verbose = true)
+int CompareResults(CUB_NS_QUALIFIER::NullType* computed, CUB_NS_QUALIFIER::NullType* reference, OffsetT len, bool verbose = true)
 {
     return 0;
 }
@@ -1351,8 +1481,8 @@ int CompareResults(double* computed, double* reference, OffsetT len, bool verbos
  * of a host array
  */
 int CompareDeviceResults(
-    cub::NullType */* h_reference */,
-    cub::NullType */* d_data */,
+    CUB_NS_QUALIFIER::NullType */* h_reference */,
+    CUB_NS_QUALIFIER::NullType */* d_data */,
     std::size_t /* num_items */,
     bool /* verbose */ = true,
     bool /* display_data */ = false)
@@ -1367,7 +1497,7 @@ int CompareDeviceResults(
 template <typename S, typename OffsetT>
 int CompareDeviceResults(
     S *h_reference,
-    cub::DiscardOutputIterator<OffsetT> d_data,
+    CUB_NS_QUALIFIER::DiscardOutputIterator<OffsetT> d_data,
     std::size_t num_items,
     bool verbose = true,
     bool display_data = false)
@@ -1469,7 +1599,7 @@ int CompareDeviceDeviceResults(
  * Print the contents of a host array
  */
 void DisplayResults(
-    cub::NullType   */* h_data */,
+    CUB_NS_QUALIFIER::NullType   */* h_data */,
     std::size_t      /* num_items */)
 {}
 
@@ -1528,7 +1658,7 @@ void InitializeSegments(
     if (num_segments <= 0)
         return;
 
-    unsigned int expected_segment_length = (num_items + num_segments - 1) / num_segments;
+    unsigned int expected_segment_length = CUB_NS_QUALIFIER::DivideAndRoundUp(num_items, num_segments);
     int offset = 0;
     for (int i = 0; i < num_segments; ++i)
     {
@@ -1645,3 +1775,77 @@ struct GpuTimer
         return elapsed;
     }
 };
+
+struct HugeDataType
+{
+  static constexpr int ELEMENTS_PER_OBJECT = 128;
+
+  __device__ __host__ HugeDataType()
+  {
+    for (int i = 0; i < ELEMENTS_PER_OBJECT; i++)
+    {
+      data[i] = 0;
+    }
+  }
+
+  __device__ __host__ HugeDataType(const HugeDataType&rhs)
+  {
+    for (int i = 0; i < ELEMENTS_PER_OBJECT; i++)
+    {
+      data[i] = rhs.data[i];
+    }
+  }
+
+  explicit __device__ __host__ HugeDataType(int val)
+  {
+    for (int i = 0; i < ELEMENTS_PER_OBJECT; i++)
+    {
+      data[i] = val;
+    }
+  }
+
+  int data[ELEMENTS_PER_OBJECT];
+};
+
+inline __device__ __host__ bool operator==(const HugeDataType &lhs,
+                                           const HugeDataType &rhs)
+{
+  for (int i = 0; i < HugeDataType::ELEMENTS_PER_OBJECT; i++)
+  {
+    if (lhs.data[i] != rhs.data[i])
+    {
+      return false;
+    }
+  }
+
+  return true;
+}
+
+inline __device__ __host__ bool operator<(const HugeDataType &lhs,
+                                          const HugeDataType &rhs)
+{
+  for (int i = 0; i < HugeDataType::ELEMENTS_PER_OBJECT; i++)
+  {
+    if (lhs.data[i] < rhs.data[i])
+    {
+      return true;
+    }
+  }
+
+  return false;
+}
+
+template <typename DataType>
+__device__ __host__ bool operator!=(const HugeDataType &lhs,
+                                    const DataType &rhs)
+{
+  for (int i = 0; i < HugeDataType::ELEMENTS_PER_OBJECT; i++)
+  {
+    if (lhs.data[i] != rhs)
+    {
+      return true;
+    }
+  }
+
+  return false;
+}
